"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""

import tempfile

from django.test import TestCase
from django.test.utils import override_settings

from . import models, app_settings, notification

MEDIA_ROOT = tempfile.mkdtemp()

class MailingListTreeTestCase(TestCase):
    '''Test mailing lists member resolution in a classical tree setup.'''
    def setUp(self):
        self.users = []
        self.mls = []
        for i in range(20):
            self.users.append(models.User.objects.create(username='%s' % i))
        sublist = []
        for i in range(19, -1, -1):
            self.mls.insert(0, models.MailingList.objects.create(
                    name='%s' % i))
            self.mls[0].members = [self.users[i]]
            self.mls[0].mailing_list_members = sublist
            sublist = [ self.mls[0] ]

    def test_mailing_list_recursive(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        for i in range(19, -1, -1):
            members = self.mls[i].recursive_members()
            self.assertEqual(len(members), 20-i)

    def test_mailing_list_recursive_with_origin(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        for i in range(19, -1, -1):
            mwo = self.mls[i].recursive_members_with_origin()
            self.assertEqual(len(mwo), 20-i)
            for j in range(i, 20):
                self.assertEqual(mwo[self.users[j]], set(self.mls[i:j+1]))


class MailingListCycle(TestCase):
    '''Test mailing lists member resolution in a cyclic setup.'''
    def setUp(self):
        self.users = []
        self.mls = []
        for i in range(20):
            self.users.append(models.User.objects.create(username='%s' % i))
        sublist = []
        for i in range(19, -1, -1):
            self.mls.insert(0, models.MailingList.objects.create(
                    name='%s' % i))
            self.mls[0].members = [self.users[i]]
            self.mls[0].mailing_list_members = sublist
            sublist = [ self.mls[0] ]
        self.mls[19].mailing_list_members = [ self.mls[0] ]

    def test_mailing_list_recursive(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        for i in range(19, -1, -1):
            members = self.mls[i].recursive_members()
            self.assertEqual(len(members), 20)

    def test_mailing_list_recursive_with_origin(self):
        """
        Tests that 1 + 1 always equals 2.
        """
        for i in range(19, -1, -1):
            mwo = self.mls[i].recursive_members_with_origin()
            self.assertEqual(len(mwo), 20)
            for j in range(i, 20):
                self.assertEqual(mwo[self.users[j]], set(self.mls))


@override_settings(MEDIA_ROOT=MEDIA_ROOT)
class BaseTestCase(TestCase):
        COUNT = 10

        def setUp(self):
            self.setUpUsers()
            self.setUpDocuments()

        def setUpUsers(self):
            from django.contrib.auth.models import User
            from .models import (DocbowProfile, FileType,)

            self.users = []
            for i in range(self.COUNT):
                self.users.append(
                        User.objects.create(username='user-%s' % i, 
                            email='user-%s@example.com' % i))
                self.users[-1].set_password('password')
                self.users[-1].save()
                DocbowProfile.objects.create(user=self.users[-1],
                    personal_email='personal-email-user-%s@example.com' % i)
            self.filetypes = []
            for i in range(self.COUNT):
                self.filetypes.append(
                    FileType.objects.create(name='filetype-%s' % i))

        def setUpDocuments(self):
            from django.core.files.base import ContentFile
            from .models import (Document, AttachedFile)

            self.documents = []
            for i in range(self.COUNT):
                self.documents.append(
                    Document.objects.create(sender=self.users[(i+2) % self.COUNT],
                        filetype=self.filetypes[i]))
                self.documents[-1].to_user = [self.users[i % self.COUNT], 
                        self.users[(i+1) % self.COUNT]]
                for j in range(2):
                    attached_file = AttachedFile(name='file%s' % j,
                            document=self.documents[-1],
                            kind=None)
                    attached_file.content.save('file%s.pdf' % j, ContentFile('coucou'))
                    attached_file.save()
                self.documents[-1].post()

class BasicTestCase(BaseTestCase):
    def test_notification_mail(self):
        from django.core import mail
        from django.core import management
        import re

        management.call_command('notify')
        self.assertEquals(len(mail.outbox), 20)
        outbox = sorted(mail.outbox, key=lambda m: tuple(sorted(m.to)))
        MAIL_LINK_RE = re.compile('https?://[^/]+/inbox/\d+/')
        for message, i in zip(outbox, range(20)):
            recipient = self.users[i / 2]
            if app_settings.PERSONAL_EMAIL:
                emails = [recipient.docbowprofile.personal_email, recipient.email]
            else:
                emails = [recipient.email]
            self.assertEquals(set(message.to), set(emails))
            match = MAIL_LINK_RE.search(message.body)
            self.assertIsNotNone(match)

class UtilsTestCase(BaseTestCase):
        def setUp(self):
            super(UtilsTestCase, self).setUp()
            self.user1, self.user2 = self.users[:2]
            self.user2.docbowprofile.is_guest = True
            self.user2.docbowprofile.save()

        def test_generate_filename(self):
            from . import models
            self.assertNotEquals(models.generate_filename(None, 'xxx.pdf'),
                models.generate_filename(None, 'xxx.pdf'))

        def test_all_emails(self):
            from . import models
            self.assertEquals(set(models.all_emails(self.user1)),
                set((self.user1.email, self.user1.docbowprofile.personal_email)))

        def test_document_manager(self):
            from . import models
            with self.assertNumQueries(2):
                list(models.Document.objects.all())

        def test_document_accessors(self):
            document = self.documents[0]
            self.assertEqual(document.filenames(),
                    'file0.pdf, file1.pdf')
            self.assertEqual(set(document.user_human_to()),
                    set(['user-0', 'user-1']))
            self.assertEqual(document.group_human_to(), [])
            self.assertEqual(set(document.human_to()),
                    set(['user-0', 'user-1']))
            self.assertIsNotNone(document.filename_links())
            self.assertEquals(set(document.to()), set((self.user1, self.user2)))
            self.assertEquals(dict(document.to_with_origin()),
                    {self.user1: set([ '--direct--' ]), self.user2: set(['--direct--'])})
            self.assertEqual(sorted(map(lambda x: x.pk, document.delivered_to())),
                    sorted([ self.user1.pk, self.user2.pk]))

        def test_is_guest(self):
            from . import models
            self.assertTrue(models.is_guest(self.user2))

@override_settings(AUTHENTICATION_BACKENDS=['django.contrib.auth.backends.ModelBackend'])
class DelegatesTestCase(BaseTestCase):
        def setUp(self):
            from django.contrib.auth.models import User
            from .models import Delegation

            self.setUpUsers()
            self.delegate = delegate = User(username='delegate')
            delegate.set_password('delegate')
            delegate.save()
            Delegation.objects.create(by=self.users[0], to=delegate)
            Delegation.objects.create(by=self.users[1], to=delegate)

        def test_inbox_by_document(self):
            from .models import (Document, AttachedFile)
            from django.core.files.base import ContentFile
            from django.test import Client

            document = Document.objects.create(sender=self.users[0], filetype=self.filetypes[0])
            document.to_user = [self.users[1]]
            attached_file = AttachedFile(name='file-private-flag.pdf', document=document, kind=None)
            attached_file.content.save('file-private-flag.pdf', ContentFile('coucou'))
            attached_file.save()
            document.post()
            c = Client()
            c.login(username='delegate', password='delegate')
            response = c.get('/inbox/%s/' % document.pk)
            self.assertEqual(response.status_code, 200)

        def test_private_flag(self):
            from .models import (Document, AttachedFile, Mailbox, Notification)
            from django.core.files.base import ContentFile
            from django.test import Client

            document = Document.objects.create(sender=self.users[0], filetype=self.filetypes[0], private=True)
            document.to_user = [self.users[1]]
            attached_file = AttachedFile(name='file-private-flag.pdf', document=document, kind=None)
            attached_file.content.save('file-private-flag.pdf', ContentFile('coucou'))
            attached_file.save()
            document.post()
            self.assertEqual(Mailbox.objects.count(), 2)
            c = Client()
            self.assertEqual(self.users[0].username, 'user-0')
            self.assertEqual(self.users[1].username, 'user-1')
            c.login(username='user-0', password='password')
            response = c.get('/outbox/%s/' % document.pk)
            self.assertEqual(response.status_code, 200, msg='status_code != 200 == %r content %r' % (response.status_code, response.content))
            c.login(username='user-1', password='password')
            response = c.get('/inbox/%s/' % document.pk)
            self.assertEqual(response.status_code, 200)
            c = Client()
            c.login(username='delegate', password='delegate')
            response = c.get('/inbox/%s/' % document.pk)
            self.assertEqual(response.status_code, 302)
            response = c.get('/outbox/%s/' % document.pk)
            self.assertEqual(response.status_code, 302)
            # check notifications
            with self.assertRaises(Notification.DoesNotExist):
                Notification.objects.get(user=self.delegate, document=document)
            Notification.objects.get(user=self.users[1], document=document)


class DummyNotifier(notification.BaseNotifier):
    notifications = []
    description = 'Dummy'
    key = 'dummy'

    def process(self, notification):
        self.notifications.append(notification)

    @classmethod
    def reset(cls):
        cls.notifications = []


@override_settings(DOCBOW_NOTIFIERS=('docbow_project.docbow.tests.DummyNotifier',))
class NotificationTestCase(BaseTestCase):
    COUNT = 2

    def test_notification_preferences(self):
        models.NotificationPreference.objects.create(
                user=self.users[0],
                kind=DummyNotifier.key,
                filetype=self.filetypes[0],
                value=False)
        models.NotificationPreference.objects.create(
                user=self.users[1],
                kind=DummyNotifier.key,
                filetype=self.filetypes[1],
                value=False)
        DummyNotifier.notifications = []
        notification.process_notifications()
        self.assertEqual(len(DummyNotifier.notifications), 2)
        for notif in DummyNotifier.notifications:
            with self.assertRaises(models.NotificationPreference.DoesNotExist):
                models.NotificationPreference.objects.get(
                        user=notif.user,
                        filetype=notif.document.filetype,
                        kind=DummyNotifier.key,
                        value=False)

    def test_notification_preferences(self):
        models.NotificationPreference.objects.create(
                user=self.users[0],
                kind=DummyNotifier.key,
                filetype=self.filetypes[0],
                value=False)
        models.NotificationPreference.objects.create(
                user=self.users[1],
                kind=DummyNotifier.key,
                filetype=self.filetypes[1],
                value=False)
        DummyNotifier.notifications = []
        notification.process_notifications()
        self.assertEqual(len(DummyNotifier.notifications), 2)
        for notif in DummyNotifier.notifications:
            with self.assertRaises(models.NotificationPreference.DoesNotExist):
                models.NotificationPreference.objects.get(
                        user=notif.user,
                        filetype=notif.document.filetype,
                        kind=DummyNotifier.key,
                        value=False)

@override_settings(DOCBOW_NOTIFIERS=('docbow_project.docbow.tests.DummyNotifier',), MEDIA_ROOT=MEDIA_ROOT)
class NotificationToDelegatesTestCase(TestCase): 
    def send_document(self, sender, user_recipients=[], list_recipients=[],
            filetype_name='dummy filetype',
            names_and_contents=(('dummy.pdf', 'dummy content'),)):
        from django.core.files.base import ContentFile
        from docbow_project.docbow.models import (FileType, Document, AttachedFile)

        filetype, created = FileType.objects.get_or_create(name=filetype_name)
        document = Document.objects.create(sender=sender, filetype=filetype)
        document.to_user = user_recipients
        document.to_list = list_recipients
        attached_files = []
        for name, content in names_and_contents:
            attached_file = AttachedFile(name=name,
                    document=document, kind=None)
            attached_file.content.save(name, ContentFile(content))
            attached_file.save()
            attached_files.append(attached_file)

        document.post()
        return document, attached_files

    def test_notification_to_delegates(self):
        from django.contrib.auth.models import User
        from docbow_project.docbow.models import Delegation

        DummyNotifier.reset()
        sender = User.objects.create(username='sender')
        recipient = User.objects.create(username='recipient')
        delegate = User.objects.create(username='delegate')
        Delegation.objects.create(by=recipient, to=delegate)
        self.send_document(sender, [recipient])

        notification.process_notifications()
        self.assertEqual(len(DummyNotifier.notifications), 2)
        self.assertEqual(set([notif.user for notif in
            DummyNotifier.notifications]), set([recipient, delegate]))

        delegate.is_active = False
        delegate.save()
        DummyNotifier.reset()
        self.send_document(sender, [recipient])
        notification.process_notifications()
        self.assertEqual(len(DummyNotifier.notifications), 1)
        self.assertEqual(DummyNotifier.notifications[0].user, recipient)

